package de.felix.portfolio;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.PriorityQueue;
import java.util.stream.DoubleStream;

public class Main {
	public static void main(String[] args) {
		/*
		 * Here is a small example of usage:
		 */
		
		double rebalancingCapital = 300;
		
		double[] currentAmount    = {100, 80, 120, 80};
		double[] targetAllocation = {0.1, 0.1, 0.6, 0.2};
		double[] minBuyAmount     = {25, 50, 0, -10};
		
		double[] invests = rebalance(rebalancingCapital, currentAmount, targetAllocation, minBuyAmount);
		
		System.out.println(Arrays.toString(invests));	// [25.0, 50.0, 228.5, -3.499999999999986]
		
		double deviation[] = new double[invests.length];
	    Arrays.setAll(deviation, i -> (currentAmount[i] + invests[i]) / (DoubleStream.of(currentAmount).sum() + rebalancingCapital) - targetAllocation[i]);

		System.out.println(Arrays.toString(deviation));	// [0.08382352941176471, 0.09117647058823528, -0.08750000000000002, -0.0875]
		
		/*
		 * deviation-Array shows, that the first and second asset take a bigger part of the portfolio than desired
		 * invests-Array shows, that the first and second asset were only bought as much as necessary (minBuyAmount)
		 * 
		 * deviation-Array shows furthermore, that the third and fourth asset do have the same deviation after rebalancing (which minimizes the squared sum)
		 * 
		 * invests-Array shows furthermore, that the fourth asset will partially be sold in order to minimize the error (this was allowed by the negative minBuyAmount constraint)
		 */
	}
	
	/**
	 * Rebalances the portfolio with additional capital under certain constraints. It minimizes the Sum Squared Error between 
	 * the actual asset allocation after rebalancing and the desired asset allocation.
	 * 
	 * @param capital Additional capital which shall be used for rebalancing. 
	 * @param currentAmount Amounts of invested money per asset before the rebalancing.
	 * @param targetAllocation Desired allocation per asset in percent.
	 * @param minBuyAmount Minimum amount of money which has to be invested per asset. Negative values are possible. 
	 * This allows the algorithm to reduce this asset (e. g. cash) in order to minimize the error.
	 * @return Returns the amount of money per asset which should be invested in order to minimize the squared deviation.
	 */
	public static double[] rebalance(double capital, double[] currentAmount, double[] targetAllocation, double[] minBuyAmount) {
		
		assert(currentAmount.length >= 1);
		assert(currentAmount.length == targetAllocation.length);
		assert(currentAmount.length == minBuyAmount.length);
		assert(DoubleStream.of(minBuyAmount).sum() < capital);
		assert(Math.abs(DoubleStream.of(targetAllocation).sum() - 1) < 1e-6);
		
		double totalCapital = DoubleStream.of(currentAmount).sum() + capital;
		
		// initialize the optimization variable (invests) in order to fulfill the minimum buy constraints
		double[] invests = minBuyAmount.clone();
		capital -= DoubleStream.of(minBuyAmount).sum();
		
		// calculate the squared deviation per asset and store it with associated index
		PriorityQueue<Entry> deltas = new PriorityQueue<Entry>();

		for (int i = 0; i < currentAmount.length; i++) {
			double delta = (currentAmount[i] + invests[i]) / totalCapital - targetAllocation[i]; 
			deltas.add(new Entry(delta, i));
		}
		
		// start iteration with the asset which has the biggest squared deviation
		List<Entry> balancedAssets = new ArrayList<Entry>();
		balancedAssets.add(deltas.poll());		
		
		// iterate until every asset was considered or the rebalancing capital is completely distributed
		while (!deltas.isEmpty() && capital > 0) {
			// get the asset with the next biggest squared deviation
			Entry curAsset = deltas.poll();
			
			// calculate the amount of money which needs to be invested in balancedAssets in order to achieve the same squared deviation as curAsset
			double invest = Math.min(totalCapital * (curAsset.key - balancedAssets.get(0).key), capital / balancedAssets.size());

			// add investments to balanced assets in order to bring them to the same squared deviation as curAsset
			for (Entry e : balancedAssets) {
				invests[e.value] += invest;
			}
			capital -= invest * balancedAssets.size();
			
			// set new squared deviation and add current (newly balanced) asset
			for (Entry e : balancedAssets) {
				e.key = curAsset.key;
			}
			balancedAssets.add(curAsset);	
		}
		
		// distribute the remaining capital evenly
		for (int i = 0; i < invests.length; i++) {
			invests[i] += capital / invests.length;
		} 
		capital = 0;
	
		return invests;
	}
	
	
	/**
	 * Pair of values, which can be sorted by the key
	 */
	private static class Entry implements Comparable<Entry> {
	    public double key;
	    public int value;

	    public Entry(double key, int value) {
	        this.key = key;
	        this.value = value;
	    }

	    @Override
	    public int compareTo(Entry other) {
	        return ((Double) key).compareTo(other.key);
	    }
	}
	
}
